# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.

from megatron.core.inference.text_generation_controllers.text_generation_controller import (  # noqa: F401 # pylint: disable=unused-import
    TextGenerationController as SimpleTextGenerationController,
)
